import enum

import pytest
from sqlalchemy import Column, func, select, types
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import column_property, composite

try:
    from sqlalchemy_utils import ChoiceType, JSONType, ScalarListType, TSVectorType
except ImportError:
    ChoiceType = JSONType = ScalarListType = TSVectorType = object

import graphene
from graphene.relay import Node
from graphene.types.datetime import DateTime
from graphene.types.json import JSONString

from alchql.converter import (
    convert_sqlalchemy_column,
    convert_sqlalchemy_composite,
    convert_sqlalchemy_relationship,
)
from alchql.fields import (
    UnsortedSQLAlchemyConnectionField,
    default_connection_field_factory,
)
from alchql.registry import Registry, get_global_registry
from alchql.types import SQLAlchemyObjectType
from .models import Article, CompositeFullName, Pet, Reporter


def mock_resolver():
    """Empty resolver"""


def get_field(sqlalchemy_type, **column_kwargs):
    class Model(declarative_base()):
        __tablename__ = "model"
        id_ = Column(types.Integer, primary_key=True)
        column = Column(sqlalchemy_type, doc="Custom Help Text", **column_kwargs)

    column_prop = inspect(Model).column_attrs["column"]
    return convert_sqlalchemy_column(column_prop, get_global_registry(), mock_resolver)


def get_field_from_column(column_):
    class Model(declarative_base()):
        __tablename__ = "model"
        id_ = Column(types.Integer, primary_key=True)
        column = column_

    column_prop = inspect(Model).column_attrs["column"]
    return convert_sqlalchemy_column(column_prop, get_global_registry(), mock_resolver)


def test_should_unknown_sqlalchemy_field_raise_exception():
    re_err = "Don't know how to convert the SQLAlchemy field"

    # support legacy Binary type and subsequent LargeBinary

    LargeBinary = getattr(types, "LargeBinary")

    with pytest.raises(Exception, match=re_err):
        get_field(LargeBinary())


def test_should_date_convert_string():
    assert get_field(types.Date()).type == graphene.Date


def test_should_datetime_convert_datetime():
    assert get_field(types.DateTime()).type == DateTime


def test_should_time_convert_string():
    assert get_field(types.Time()).type == graphene.String


def test_should_string_convert_string():
    assert get_field(types.String()).type == graphene.String


def test_should_text_convert_string():
    assert get_field(types.Text()).type == graphene.String


def test_should_unicode_convert_string():
    assert get_field(types.Unicode()).type == graphene.String


def test_should_unicodetext_convert_string():
    assert get_field(types.UnicodeText()).type == graphene.String


def test_should_enum_convert_enum():
    field = get_field(types.Enum(enum.Enum("TwoNumbers", ("one", "two"))))
    field_type = field.type()
    assert isinstance(field_type, graphene.Enum)
    assert field_type._meta.name == "TwoNumbers"
    assert hasattr(field_type, "ONE")
    assert not hasattr(field_type, "one")
    assert hasattr(field_type, "TWO")
    assert not hasattr(field_type, "two")

    field = get_field(types.Enum("one", "two", name="two_numbers"))
    field_type = field.type()
    assert isinstance(field_type, graphene.Enum)
    assert field_type._meta.name == "TwoNumbers"
    assert hasattr(field_type, "ONE")
    assert not hasattr(field_type, "one")
    assert hasattr(field_type, "TWO")
    assert not hasattr(field_type, "two")


def test_should_not_enum_convert_enum_without_name():
    field = get_field(types.Enum("one", "two"))
    re_err = r"No type name specified for Enum\('one', 'two'\)"
    with pytest.raises(TypeError, match=re_err):
        field.type()


def test_should_small_integer_convert_int():
    assert get_field(types.SmallInteger()).type == graphene.Int


def test_should_big_integer_convert_float():
    assert get_field(types.BigInteger()).type == graphene.Float


def test_should_integer_convert_int():
    assert get_field(types.Integer()).type == graphene.Int


def test_should_primary_integer_convert_id():
    assert get_field(types.Integer(), primary_key=True).type == graphene.NonNull(
        graphene.ID
    )


def test_should_boolean_convert_boolean():
    assert get_field(types.Boolean()).type == graphene.Boolean


def test_should_float_convert_float():
    assert get_field(types.Float()).type == graphene.Float


def test_should_numeric_convert_float():
    assert get_field(types.Numeric()).type == graphene.Float


def test_should_choice_convert_enum():
    field = get_field(ChoiceType([("es", "Spanish"), ("en", "English")]))
    graphene_type = field.type
    assert issubclass(graphene_type, graphene.Enum)
    assert graphene_type._meta.name == "MODEL_COLUMN"
    assert graphene_type._meta.enum.__members__["es"].value == "Spanish"
    assert graphene_type._meta.enum.__members__["en"].value == "English"


def test_should_enum_choice_convert_enum():
    class TestEnum(enum.Enum):
        es = "Spanish"
        en = "English"

    field = get_field(ChoiceType(TestEnum, impl=types.String()))
    graphene_type = field.type
    assert issubclass(graphene_type, graphene.Enum)
    assert graphene_type._meta.name == "MODEL_COLUMN"
    assert graphene_type._meta.enum.__members__["es"].value == "Spanish"
    assert graphene_type._meta.enum.__members__["en"].value == "English"


def test_should_intenum_choice_convert_enum():
    class TestEnum(enum.IntEnum):
        one = 1
        two = 2

    field = get_field(ChoiceType(TestEnum, impl=types.String()))
    graphene_type = field.type
    assert issubclass(graphene_type, graphene.Enum)
    assert graphene_type._meta.name == "MODEL_COLUMN"
    assert graphene_type._meta.enum.__members__["one"].value == 1
    assert graphene_type._meta.enum.__members__["two"].value == 2


def test_should_columproperty_convert():
    field = get_field_from_column(
        column_property(select([func.sum(func.cast(id, types.Integer))]).where(id == 1))
    )

    assert field.type == graphene.Int


def test_should_scalar_list_convert_list():
    field = get_field(ScalarListType())
    assert isinstance(field.type, graphene.List)
    assert field.type.of_type == graphene.String


def test_should_jsontype_convert_jsonstring():
    assert get_field(JSONType()).type == JSONString


def test_should_tsvector_convert_string():
    assert get_field(TSVectorType).type == graphene.String


def test_should_tsvector_convert_string2():
    assert get_field(postgresql.TSVECTOR()).type == graphene.String


def test_should_manytomany_convert_connectionorlist():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Article

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.pets.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    assert not dynamic_field.get_type()


def test_should_manytomany_convert_connectionorlist_list():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Pet

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.pets.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert isinstance(graphene_type.type, graphene.List)
    assert graphene_type.type.of_type == A


def test_should_manytomany_convert_connectionorlist_connection():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (Node,)

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.pets.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    assert isinstance(dynamic_field.get_type(), UnsortedSQLAlchemyConnectionField)


def test_should_manytoone_convert_connectionorlist():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Article

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.pets.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    assert not dynamic_field.get_type()


def test_should_manytoone_convert_connectionorlist_list():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Article

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.favorite_article.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == A


def test_should_manytoone_convert_connectionorlist_connection():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.favorite_article.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == A


def test_should_onetoone_convert_field():
    class A(SQLAlchemyObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    dynamic_field = convert_sqlalchemy_relationship(
        Reporter.favorite_article.property,
        A,
        default_connection_field_factory,
        "orm_field_name",
    )
    assert isinstance(dynamic_field, graphene.Dynamic)
    graphene_type = dynamic_field.get_type()
    assert isinstance(graphene_type, graphene.Field)
    assert graphene_type.type == A


def test_should_postgresql_uuid_convert():
    assert get_field(postgresql.UUID()).type == graphene.String


def test_should_postgresql_enum_convert():
    field = get_field(postgresql.ENUM("one", "two", name="two_numbers"))
    field_type = field.type()
    assert isinstance(field_type, graphene.Enum)
    assert field_type._meta.name == "TwoNumbers"
    assert hasattr(field_type, "ONE")
    assert not hasattr(field_type, "one")
    assert hasattr(field_type, "TWO")
    assert not hasattr(field_type, "two")


def test_should_postgresql_py_enum_convert():
    field = get_field(
        postgresql.ENUM(enum.Enum("TwoNumbers", "one two"), name="two_numbers")
    )
    field_type = field.type()
    assert field_type._meta.name == "TwoNumbers"
    assert isinstance(field_type, graphene.Enum)
    assert hasattr(field_type, "ONE")
    assert not hasattr(field_type, "one")
    assert hasattr(field_type, "TWO")
    assert not hasattr(field_type, "two")


def test_should_postgresql_array_convert():
    field = get_field(postgresql.ARRAY(types.Integer))
    assert isinstance(field.type, graphene.List)
    assert field.type.of_type == graphene.Int


def test_should_array_convert():
    field = get_field(types.ARRAY(types.Integer))
    assert isinstance(field.type, graphene.List)
    assert field.type.of_type == graphene.Int


def test_should_postgresql_json_convert():
    assert get_field(postgresql.JSON()).type == graphene.JSONString


def test_should_postgresql_jsonb_convert():
    assert get_field(postgresql.JSONB()).type == graphene.JSONString


def test_should_postgresql_hstore_convert():
    assert get_field(postgresql.HSTORE()).type == graphene.JSONString


def test_should_composite_convert():
    registry = Registry()

    class CompositeClass:
        def __init__(self, col1, col2):
            self.col1 = col1
            self.col2 = col2

    @convert_sqlalchemy_composite.register(CompositeClass, registry)
    def convert_composite_class(composite, registry):
        return graphene.String(description=composite.doc)

    field = convert_sqlalchemy_composite(
        composite(
            CompositeClass,
            (Column(types.Unicode(50)), Column(types.Unicode(50))),
            doc="Custom Help Text",
        ),
        registry,
        mock_resolver,
    )
    assert isinstance(field, graphene.String)


def test_should_unknown_sqlalchemy_composite_raise_exception():
    class CompositeClass:
        def __init__(self, col1, col2):
            self.col1 = col1
            self.col2 = col2

    re_err = "Don't know how to convert the composite field"
    with pytest.raises(Exception, match=re_err):
        convert_sqlalchemy_composite(
            composite(
                CompositeFullName,
                (Column(types.Unicode(50)), Column(types.Unicode(50))),
            ),
            Registry(),
            mock_resolver,
        )



import enum
from collections import defaultdict

import sqlalchemy as sa
from aiodataloader import DataLoader
from graphene import ResolveInfo
from sqlalchemy import ForeignKey
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeMeta, RelationshipProperty
from sqlalchemy.sql import Select

from .query_helper import QueryHelper
from .utils import EnumValue, filter_requested_fields_for_object, table_to_class


def get_join(relation: RelationshipProperty):
    if relation.primaryjoin is not None:
        l = relation.primaryjoin.left.table
        r = relation.primaryjoin.right.table

        base_table = l
        join_table = r
        join_table2 = None

        if relation.secondaryjoin is not None:
            l2 = relation.secondaryjoin.right.table
            r2 = relation.secondaryjoin.left.table
            if l == l2:
                base_table = l
                join_table = r
                join_table2 = r2
            elif r == l2:
                base_table = r
                join_table = l
                join_table2 = r2
            elif r == r2:
                base_table = r
                join_table = l
                join_table2 = l2
            elif l == r2:
                base_table = l
                join_table = r
                join_table2 = l2

        join = sa.join(
            base_table,
            join_table,
            relation.primaryjoin,
        )

        if join_table2 is not None:
            join = join.join(
                join_table2,
                relation.secondaryjoin,
            )

        return join


class BaseLoader(DataLoader):
    target_field: sa.Column
    target: DeclarativeMeta

    def __init__(
        self,
        session: AsyncSession,
        info: ResolveInfo = None,
        *args,
        **kwargs,
    ):
        self.session = session
        self.info = info
        self.fields = set()

        super().__init__(*args, **kwargs)

    @staticmethod
    def get_sort_args(gql_field, object_type):
        sort_args = []

        if gql_field.arguments.get("sort") is not None:
            sort = gql_field.arguments["sort"]
            if not isinstance(sort, list):
                sort = [sort]

            if hasattr(object_type, "sort_argument"):
                sort_type = object_type.sort_argument().type.of_type
                new_sort = []
                for s in sort:
                    if isinstance(s, (EnumValue, enum.Enum)):
                        new_sort.append(s.value)
                    else:
                        new_sort.append(getattr(sort_type, s).value)

                sort = new_sort

            for item in sort:
                sort_args.append(item)

        return sort_args

    def prepare_query(self, q: Select) -> Select:
        return q

    async def batch_load_fn(self, keys):
        object_types = getattr(self.info.context, "object_types", {})
        object_type = object_types.get(self.info.field_name)

        if object_type is None:
            parent_type = self.info.parent_type
            field = parent_type.fields[self.info.field_name]
            object_type = field.type.graphene_type

        filters = QueryHelper.get_filters(self.info)
        gql_field = QueryHelper.get_current_field(self.info)
        sort_args = self.get_sort_args(gql_field, object_type)

        selected_fields = QueryHelper.get_selected_fields(
            self.info, model=self.target, object_type=object_type, sort=sort_args
        )
        if not selected_fields:
            selected_fields = self.fields or self.target.__table__.columns

        q = (
            sa.select(
                *selected_fields,
                self.target_field.label("_batch_key"),
            )
            .where(self.target_field.in_(keys))
            .order_by(*sort_args)
        )
        q = self.prepare_query(q)

        if object_type and hasattr(object_type, "set_select_from"):
            setattr(self.info.context, "keys", keys)
            q = await object_type.set_select_from(self.info, q, gql_field.values)
            if list(q._group_by_clause):
                q = q.group_by(self.target_field)

        if filters:
            q = q.where(sa.and_(*filters))

        results_by_ids = defaultdict(list)

        conversion_type = object_type or self.target
        results = map(dict, await self.session.execute(q))

        for result in results:
            _batch_key = result.pop("_batch_key")
            _data = filter_requested_fields_for_object(result, conversion_type)
            results_by_ids[_batch_key].append(conversion_type(**_data))

        return [results_by_ids.get(key, []) for key in keys]


def generate_loader_by_relationship(relation: RelationshipProperty):
    _target_field = next(iter(relation.local_columns))
    _target = relation.mapper.entity

    class RelationLoader(BaseLoader):
        target = _target
        target_field = _target_field

        def prepare_query(self, q: Select) -> Select:
            join = get_join(relation)
            if join is not None:
                q = q.select_from(join)
            if relation.order_by:
                for n, ob in enumerate(relation.order_by):
                    q = q.order_by(ob)
                    """fix for bug: for SELECT DISTINCT, ORDER BY expressions must appear in select list"""
                    q = q.add_columns(ob.label(f"order_by_{n}"))
            q = q.distinct()

            return q

    return RelationLoader


def generate_loader_by_foreign_key(fk: ForeignKey, reverse=False):
    if not reverse:
        _target_field = fk.column
        _target = table_to_class(_target_field.table)
    else:
        _target_field = fk.parent
        _target = table_to_class(_target_field.table)

    class FkLoader(BaseLoader):
        target = _target
        target_field = _target_field

    return FkLoader



from unittest import mock

import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from alchql.connection import from_query
from alchql.consts import OP_ILIKE
from alchql.fields import FilterConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import SQLAlchemyObjectType
from .models import Editor


async def add_test_data(session):
    await session.execute(
        sa.insert(Editor).values(
            [
                {
                    Editor.name: f"Editor#{num}",
                }
                for num in range(100)
            ]
        )
    )


class CountableConnectionCreator:
    @classmethod
    def create_type(cls, connection_name, **kwargs):
        class CountableConnection(graphene.relay.Connection):
            count = graphene.Int()
            total_count = graphene.Int()

            class Meta:
                name = connection_name
                node = kwargs["node"]

            @staticmethod
            def resolve_count(root, info, **args):
                return len(root.edges)

        return CountableConnection


async def get_query():
    class EditorType(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            interfaces = (AsyncNode,)
            connection_class = CountableConnectionCreator
            filter_fields = {
                Editor.name: [OP_ILIKE],
            }

    class Query(graphene.ObjectType):
        node = graphene.relay.Node.Field()
        editors = FilterConnectionField(EditorType, sort=EditorType.sort_argument())

    return Query


@pytest.mark.asyncio
async def test_counters(session, raise_graphql):
    await add_test_data(session)

    query = """
    query {
      editors {
        count
        totalCount
      }
    }
    """

    schema = graphene.Schema(query=await get_query())

    with mock.patch.object(from_query, "DEFAULT_LIMIT", 50):
        result = await schema.execute_async(
            query,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Editor]),
            ],
        )

    assert not result.errors
    assert result.data == {"editors": {"count": 50, "totalCount": 100}}



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context
from unittest import mock

from .models import Editor
from alchql.consts import OP_ILIKE
from alchql.types import SQLAlchemyObjectType
from alchql.fields import FilterConnectionField
from alchql.node import AsyncNode
from alchql.connection import from_query
from alchql.middlewares import LoaderMiddleware


async def add_test_data(session):
    await session.execute(
        sa.insert(Editor).values(
            [
                {
                    Editor.name: f"Editor#{num}",
                }
                for num in range(100)
            ]
        )
    )


async def get_query():
    class EditorType(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            interfaces = (AsyncNode,)
            filter_fields = {
                Editor.name: [OP_ILIKE],
            }

    class Query(graphene.ObjectType):
        node = graphene.relay.Node.Field()
        editors = FilterConnectionField(EditorType, sort=EditorType.sort_argument())

    return Query


async def get_start_end_cursor(first: int, session):
    query = (
        """
            query {
              editors (first: %s) {
                edges {
                  node {
                    id
                    name
                  }
                }
                pageInfo {
                  startCursor
                  endCursor
                  hasPreviousPage
                  hasNextPage
                }
              }
            }
        """
        % first
    )

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )
    page_info = result.data["editors"]["pageInfo"]
    start_cursor = page_info["startCursor"]
    end_cursor = page_info["endCursor"]
    return start_cursor, end_cursor


@pytest.mark.asyncio
async def test_query_no_filters(session):
    await add_test_data(session)

    query = """
            query {
              editors {
                edges {
                  node {
                    id
                    name
                  }
                }
                pageInfo {
                  startCursor
                  endCursor
                  hasPreviousPage
                  hasNextPage
                }
              }
            }
        """

    limit = 10
    schema = graphene.Schema(query=await get_query())
    with mock.patch.object(from_query, "DEFAULT_LIMIT", limit):
        result = await schema.execute_async(
            query,
            context_value=Context(session=session),
            middleware=[
                LoaderMiddleware([Editor]),
            ],
        )
        assert not result.errors
        assert len(result.data["editors"]["edges"]) == limit

        page_info = result.data["editors"]["pageInfo"]
        assert page_info["startCursor"]
        assert page_info["endCursor"]
        assert not page_info["hasPreviousPage"]
        assert page_info["hasNextPage"]


@pytest.mark.asyncio
async def test_query_first_specified(session):
    await add_test_data(session)

    first = 50

    query = (
        """
            query {
              editors (first: %s) {
                edges {
                  node {
                    id
                    name
                  }
                }
                pageInfo {
                  startCursor
                  endCursor
                  hasPreviousPage
                  hasNextPage
                }
              }
            }
        """
        % first
    )

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )
    assert len(result.data["editors"]["edges"]) == first

    page_info = result.data["editors"]["pageInfo"]
    assert page_info["startCursor"]
    assert page_info["endCursor"]
    assert not page_info["hasPreviousPage"]
    assert page_info["hasNextPage"]


@pytest.mark.asyncio
async def test_query_first_after_specified(session):
    await add_test_data(session)

    _, end_cursor = await get_start_end_cursor(10, session)

    query = (
        """
        query {
          editors (first: 10, after: "%s") {
            edges {
              node {
                id
                name
              }
            }
            pageInfo {
              startCursor
              endCursor
              hasPreviousPage
              hasNextPage
            }
          }
        }
    """
        % end_cursor
    )

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )

    editors = result.data["editors"]["edges"]
    assert len(editors) == 10
    assert editors[0]["node"]["name"] == "Editor#10"

    page_info = result.data["editors"]["pageInfo"]
    assert page_info["startCursor"]
    assert page_info["endCursor"]
    assert page_info["hasPreviousPage"]
    assert page_info["hasNextPage"]


@pytest.mark.asyncio
async def test_last_before_specified(session):
    await add_test_data(session)

    _, end_cursor = await get_start_end_cursor(20, session)

    query = (
        """
        query {
          editors (
            last: 10, 
            name_Ilike: "Editor", 
            before: "%s"
          ) {
            edges {
              node {
                id
                name
              }
            }
            pageInfo {
              startCursor
              endCursor
              hasPreviousPage
              hasNextPage
            }
          }
        }
    """
        % end_cursor
    )

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )

    editors = result.data["editors"]["edges"]
    assert len(editors) == 10
    assert editors[0]["node"]["name"] == "Editor#9"

    page_info = result.data["editors"]["pageInfo"]
    assert page_info["startCursor"]
    assert page_info["endCursor"]
    assert page_info["hasPreviousPage"]
    assert page_info["hasNextPage"]



from typing import Callable, Dict, Tuple

OP_LTE = "lte"
OP_EQ = "eq"
OP_GTE = "gte"
OP_ILIKE = "ilike"
OP_IN = "in"
OP_EMPTY = "empty"
OP_NOT_EMPTY = "not_empty"
OP_CONTAINS = "contains"
OP_LT = "lt"
OP_GT = "gt"

OPERATORS_MAPPING: Dict[str, Tuple[str, Callable]] = {
    OP_LTE: ("__le__", lambda v: v),
    OP_EQ: ("__eq__", lambda v: v),
    OP_GTE: ("__ge__", lambda v: v),
    OP_ILIKE: ("ilike", lambda v: f"%{v}%"),
    OP_IN: ("in_", lambda v: v),
    OP_EMPTY: ("is_", lambda v: None),
    OP_NOT_EMPTY: ("isnot", lambda v: None),
    OP_CONTAINS: ("contains", lambda v: v),
    OP_LT: ("__lt__", lambda v: v),
    OP_GT: ("__gt__", lambda v: v),
}



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context

from alchql.consts import OP_ILIKE
from alchql.fields import FilterConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import SQLAlchemyObjectType
from .models import Editor


async def add_test_data(session):
    await session.execute(
        sa.insert(Editor).values(
            [
                {
                    Editor.name: f"Editor#{num}",
                }
                for num in range(100)
            ]
        )
    )


async def get_query():
    class EditorType(SQLAlchemyObjectType):
        class Meta:
            model = Editor
            interfaces = (AsyncNode,)
            filter_fields = {
                Editor.name: [OP_ILIKE],
            }

    class Query(graphene.ObjectType):
        node = graphene.relay.Node.Field()
        editors = FilterConnectionField(EditorType, sort=EditorType.sort_argument())

    return Query


@pytest.mark.asyncio
async def test_query_page_info_full(session, raise_graphql):
    await add_test_data(session)

    query = """
    query {
      editors {
        pageInfo {
          startCursor
          endCursor
          hasPreviousPage
          hasNextPage
        }
      }
    }
    """

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )
    assert not result.errors


@pytest.mark.asyncio
async def test_query_page_info(session, raise_graphql):
    await add_test_data(session)

    query = """
    query {
      editors {
        pageInfo {
          startCursor
          hasNextPage
        }
      }
    }
    """

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )
    assert not result.errors



import base64
import json
from typing import NamedTuple, Type, TypeVar, Union

_T = TypeVar("_T")


class ResolvedGlobalId(NamedTuple):
    type: str
    id: Union[str, int]

    def encode(self) -> str:
        text = f"{self.type}:{json.dumps(self.id)}"
        return base64.b64encode(text.encode()).decode()

    @classmethod
    def decode(cls: Type[_T], encoded_id: str) -> _T:
        if not encoded_id:
            raise ValueError("Empty ID")

        try:
            text = base64.b64decode(encoded_id).decode()
        except Exception as e:
            raise ValueError("Invalid base64")

        type_name, id_text = text.split(":")

        return cls(
            type=type_name,
            id=json.loads(id_text),
        )

    def __str__(self):
        return self.encode()



import logging
from typing import Optional, Type

import sqlalchemy as sa
from graphene import Connection, PageInfo
from graphene.types import ResolveInfo
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.sql import Select

from .utils import (
    get_offset_with_default,
    offset_to_cursor,
)
from ..query_helper import QueryHelper
from ..utils import filter_requested_fields_for_object

DEFAULT_LIMIT = 1000


def get_count_query(query: Select, model):
    only_q = query.with_only_columns(
        *sa.inspect(model).primary_key,
    ).order_by(None)
    return sa.select(sa.func.count()).select_from(only_q.alias())


def construct_page_info(
    cls: Type[PageInfo],
    info: ResolveInfo,
    edges: list,
    limit: int,
    offset: int,
) -> PageInfo:
    page_info_kwargs = {}

    page_info_fields = QueryHelper.get_page_info_fields(info)

    if not page_info_fields:
        return cls()

    for field in page_info_fields:
        if field == "has_previous_page":
            page_info_kwargs[field] = offset > 0
        elif field == "has_next_page":
            page_info_kwargs[field] = limit and len(edges) > limit
        elif edges:
            if field == "start_cursor":
                page_info_kwargs[field] = edges[0].cursor
            elif field == "end_cursor":
                page_info_kwargs[field] = edges[:limit][-1].cursor

    return cls(**page_info_kwargs)


async def connection_from_query(
    query: Select,
    model: Type[DeclarativeMeta],
    info: ResolveInfo,
    args: Optional[dict] = None,
    connection_type: Type[Connection] = Connection,
) -> Connection:
    """
    Given a slice (subset) of an array, returns a connection object for use in
    GraphQL.
    This function is similar to `connectionFromArray`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire array, and instead wish pass in a slice of the
    total result large enough to cover the range specified in `args`.
    """
    args = args or {}
    session: AsyncSession = info.context.session

    # has_last_arg = QueryHelper.has_arg(info, "last")
    # if not QueryHelper.get_filters(info) and has_last_arg:
    #     raise TypeError('Cannot set "last" without filters applied')

    edge_type = connection_type.Edge
    node_type = edge_type.node.type
    page_info_type = getattr(connection_type, "PageInfo", PageInfo)

    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")

    current_field = QueryHelper.get_current_field(info)
    has_total_count = current_field and "total_count" in {
        i.name for i in current_field.values
    }

    if (before, after, first, last) == (None, None, None, None):
        first = DEFAULT_LIMIT
        logging.warning(f"Query without border, {first=}")

    total_count = None
    # TODO: Move total_count to PageInfo
    if (last and not before) or (after and not first and not before) or has_total_count:
        count_query = get_count_query(query, model)
        right_offset = total_count = (await session.execute(count_query)).scalar()
    else:
        right_offset = get_offset_with_default(before)

    left_offset = get_offset_with_default(after) + 1 if after else 0

    if isinstance(first, int):
        right_offset = min(left_offset + first, right_offset)
    if isinstance(last, int):
        left_offset = max(left_offset, right_offset - last)

    _slice = query
    # If supplied slice is too large, trim it down before mapping over it.
    limit = first or last or (right_offset - left_offset)

    if limit:
        _slice = _slice.limit(limit + 1)
    if left_offset:
        _slice = _slice.offset(left_offset)

    edges = []

    for i, v in enumerate(await session.execute(_slice)):
        node_value = filter_requested_fields_for_object(dict(v), node_type)
        edge = edge_type(
            node=node_type(**node_value),
            cursor=offset_to_cursor(left_offset + i),
        )
        edges.append(edge)

    connection = connection_type(
        edges=edges[:limit],
        page_info=construct_page_info(
            cls=page_info_type,
            info=info,
            edges=edges,
            limit=limit,
            offset=left_offset,
        ),
    )

    if has_total_count:
        connection.total_count = total_count

    return connection



from typing import Callable, TYPE_CHECKING, Type

from graphene.utils.get_unbound_function import get_unbound_function

if TYPE_CHECKING:
    from .types import SQLAlchemyObjectType


def get_custom_resolver(obj_type: Type["SQLAlchemyObjectType"], orm_field_name):
    """
    Since `graphene` will call `resolve_<field_name>` on a field only if it
    does not have a `resolver`, we need to re-implement that logic here so
    users are able to override the default resolvers that we provide.
    """
    resolver = getattr(obj_type, f"resolve_{orm_field_name}", None)
    if resolver:
        return get_unbound_function(resolver)

    return None


def get_attr_resolver(model_attr: str) -> Callable:
    """
    In order to support field renaming via `ORMField.model_attr`,
    we need to define resolver functions for each field.
    """

    async def resolver(root, _info):
        return getattr(root, model_attr, None)

    return resolver



import graphene
import pytest
import sqlalchemy as sa
from graphene import Context
from sqlalchemy.ext.asyncio import AsyncSession

from alchql.fields import FilterConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.types import SQLAlchemyObjectType
from .models import Editor, Reporter


async def add_test_data(session: AsyncSession):
    await session.execute(
        sa.insert(Reporter).values(
            [
                {
                    Reporter.first_name: "John",
                    Reporter.last_name: "Doe",
                    Reporter.email: "email",
                },
                {
                    Reporter.first_name: "John1",
                    Reporter.last_name: "Doe1",
                    Reporter.email: "email1",
                },
            ]
        )
    )


async def get_query():
    class ReporterType(SQLAlchemyObjectType):
        class Meta:
            model = Reporter
            interfaces = (AsyncNode,)

    class Query(graphene.ObjectType):
        node = graphene.relay.Node.Field()
        reporters = FilterConnectionField(
            ReporterType, sort=ReporterType.sort_argument()
        )

    return Query


@pytest.mark.asyncio
async def test_query_fragments(session):
    await add_test_data(session)

    query = """
    query {
      reporters {
        edges {
          node {
            firstName
            ...ReporterFragment1
          }
        }
        pageInfo {
          startCursor
          endCursor
          hasPreviousPage
          hasNextPage
        }
      }
    }
    fragment ReporterFragment1 on ReporterType {
      lastName
      ...ReporterFragment2
    }
    fragment ReporterFragment2 on ReporterType {
      email
    }
    """

    schema = graphene.Schema(query=await get_query())
    result = await schema.execute_async(
        query,
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Editor]),
        ],
    )

    assert not result.errors
    assert result.data == {
        "reporters": {
            "edges": [
                {"node": {"email": "email", "firstName": "John", "lastName": "Doe"}},
                {"node": {"email": "email1", "firstName": "John1", "lastName": "Doe1"}},
            ],
            "pageInfo": {
                "endCursor": "YXJyYXljb25uZWN0aW9uOjE=",
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
            },
        }
    }
